/* Example of playing FLIC file with the EGI player and DirectDraw
 *
 * Much of this code is concerned with setting up DirectDraw.
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <ddraw.h>
#include "eplay.h"

#define NAME            "DDrawDemo"
#define TITLE           "Direct Draw & EGI"

const IID IID_IDirectDraw  = { 0x6C14DB80,0xA733,0x11CE, { 0xA5,0x21,0x00,0x20,0xAF,0x0B,0xE5,0x60 } };

LPDIRECTDRAW        lpDD = NULL;            // DirectDraw object
LPDIRECTDRAWSURFACE lpDDSPrimary = NULL;    // DirectDraw primary surface
LPDIRECTDRAWPALETTE lpDDPalette = NULL;     // The palette

FlicAnim            *Flic;                  // The animation object


static void ReleaseAllObjects(void)
{
  if (Flic != NULL) {
    delete Flic;
    Flic = NULL;
  } /* if */
  if (lpDD != NULL) {
    if (lpDDSPrimary != NULL) {
      lpDDSPrimary->Release();
      lpDDSPrimary = NULL;
    } /* if */
    if (lpDDPalette != NULL) {
      lpDDPalette->Release();
      lpDDPalette = NULL;
    } /* if */
    lpDD->Release();
    lpDD = NULL;
  } /* if */
}

HRESULT InitFail(HWND hwnd, HRESULT hRet, LPCTSTR szError)
{
  ReleaseAllObjects();
  MessageBox(hwnd, szError, TITLE, MB_OK);
  DestroyWindow(hwnd);
  return hRet;
}

void NextFrame(void)
{
  for ( ;; ) {
    HRESULT hRet = lpDDSPrimary->Flip( NULL, 0 );
    if (hRet == DD_OK)
      return;
    if (hRet != DDERR_WASSTILLDRAWING)
      return;
  } /* for */
}

DWORD CALLBACK EGIcallback(LPFLIC /*lpFlic*/, int code, DWORD /*param*/)
{
  // use the global variable Flic (instance of the FlicAnim class) instead
  // of the passed in variable, so one can continue to use the C++ interface
  switch (code) {
  case FN_FRAME:
    if ((Flic->GetParam(FLIC_PARAM_FRAMEBITS) & FLIC_FRAME_PAL)!=NULL) {
      PALETTEENTRY pPaletteEntry[256];
      int          index;
      LPRGBQUAD    rgbq;

      // Get the colour table of the animation
      rgbq = (LPRGBQUAD)Flic->GetData(FLIC_DATA_PALETTE);
      // Convert this to the PALETTEENTRY structure
      for (index = 0; index < 256; index++) {
        pPaletteEntry[index].peFlags = PC_NOCOLLAPSE|PC_RESERVED;
        pPaletteEntry[index].peRed = rgbq[index].rgbRed;
        pPaletteEntry[index].peGreen = rgbq[index].rgbGreen;
        pPaletteEntry[index].peBlue = rgbq[index].rgbBlue;
      } /* for */
      // All entries are filled. Set them.
      lpDDPalette->SetEntries(0, 0, 256, pPaletteEntry);
    } /* if */
    NextFrame();
    break;
  } /* switch */
  return 0L;
}

long FAR PASCAL WindowProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message) {
  case WM_DESTROY:      // Clean up and close the app
    ReleaseAllObjects();
    PostQuitMessage(0);
    return 0L;

  case WM_KEYDOWN:      // Handle any non-accelerated key commands
    switch (wParam) {
    case VK_ESCAPE:
    case VK_F12:
      PostMessage(hwnd, WM_CLOSE, 0, 0);
      return 0L;
    } /* switch */
    break;

  case WM_SETCURSOR:    // Turn off the cursor
    SetCursor(NULL);
    return TRUE;

  } /* switch */
  return DefWindowProc(hwnd, message, wParam, lParam);
}

static HRESULT InitApp(HINSTANCE hInstance, int nCmdShow)
{
  HWND                hwnd;
  WNDCLASS            wc;
  DDSURFACEDESC       ddsd;
  DDSCAPS             ddscaps;
  HRESULT             hRet;
  LPDIRECTDRAW        pDD;
  LPDIRECTDRAWSURFACE lpDDSBackground;
  BOOL                okay;

  ///////////////////////////////////////////////////////////////////////////
  // Create the window
  ///////////////////////////////////////////////////////////////////////////

  // Set up and register window class
  wc.style = CS_HREDRAW | CS_VREDRAW;
  wc.lpfnWndProc = WindowProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = NULL;
  wc.hCursor = LoadCursor(NULL, IDC_ARROW);
  wc.hbrBackground = (HBRUSH )GetStockObject(LTGRAY_BRUSH);
  wc.lpszMenuName = NULL;
  wc.lpszClassName = NAME;
  RegisterClass(&wc);

  // Create a window (it is resized later)
  hwnd = CreateWindowEx(WS_EX_TOPMOST,
                        NAME, TITLE,
                        WS_POPUP,
                        0, 0, 10, 10,
                        NULL, NULL, hInstance, NULL);
  if (!hwnd)
    return FALSE;


  ///////////////////////////////////////////////////////////////////////////
  // Open the FLIC file
  ///////////////////////////////////////////////////////////////////////////
  Flic = new FlicAnim("gears.flc");
  if (Flic == NULL)
    return InitFail(hwnd, hRet, "Opening FLIC animation FAILED");


  ///////////////////////////////////////////////////////////////////////////
  // Create DirectDraw objects
  ///////////////////////////////////////////////////////////////////////////
  hRet = DirectDrawCreate(NULL, &pDD, NULL);
  if (FAILED(hRet))
    return InitFail(hwnd, hRet, "DirectDrawCreate FAILED");

  // Fetch DirectDraw interface
  hRet = pDD->QueryInterface(IID_IDirectDraw, (LPVOID *) & lpDD);
  if (FAILED(hRet))
    return InitFail(hwnd, hRet, "QueryInterface FAILED");

  // Get exclusive mode
  hRet = lpDD->SetCooperativeLevel(hwnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
  if (FAILED(hRet))
    return InitFail(hwnd, hRet, "SetCooperativeLevel FAILED");

  // Get the size and colour resolution of the FLIC, choose the best
  // resolution
  int Width = Flic->GetParam(FLIC_PARAM_WIDTH);
  int Height = Flic->GetParam(FLIC_PARAM_HEIGHT);
  int CDepth = Flic->GetParam(FLIC_PARAM_COLORDEPTH);
  if (Width <= 320 && Height <= 200) {
    Width = 320;
    Height = 200;
  } else if (Width <= 640 && Height <= 480) {
    Width = 640;
    Height = 480;
  } else if (Width <= 800 && Height <= 600) {
    Width = 800;
    Height = 600;
  } else if (Width <= 1024 && Height <= 768) {
    Width = 1024;
    Height = 768;
  } else {
    return InitFail(hwnd, hRet, "FLIC animation too large");
  } /* if */

  // Set the video mode, adjust the window size
  hRet = lpDD->SetDisplayMode(Width, Height, CDepth);
  if (FAILED(hRet))
    return InitFail(hwnd, hRet, "SetDisplayMode FAILED");
  MoveWindow(hwnd, 0, 0, Width, Height, FALSE);

  // Center the animation on the screen
  Flic->SetParam(FLIC_PARAM_XPOS, (Width - Flic->GetParam(FLIC_PARAM_WIDTH)) / 2);
  Flic->SetParam(FLIC_PARAM_YPOS, (Height - Flic->GetParam(FLIC_PARAM_HEIGHT)) / 2);

  // Create the primary surface
  ZeroMemory(&ddsd, sizeof(ddsd));
  ddsd.dwSize = sizeof(ddsd);
  ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
  ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
  ddsd.dwBackBufferCount = 1;
  hRet = lpDD->CreateSurface(&ddsd, &lpDDSPrimary, NULL);
  if (FAILED(hRet))
    return InitFail(hwnd, hRet, "CreateSurface FAILED");

  // Get a background surface
  ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
  hRet = lpDDSPrimary->GetAttachedSurface(&ddscaps, &lpDDSBackground);
  if (FAILED(hRet))
    return InitFail(hwnd, hRet, "Retrieving secondary surface FAILED");

  // Create the palette.
  // At this point, the FLIC player does not yet have a colour table because
  // it has not yet read a frame. So what we do here is to create a "black"
  // palette, and have the correct colours set in the callback.
  if (Flic->GetParam(FLIC_PARAM_COLORDEPTH) == 8) {
    PALETTEENTRY        pPaletteEntry[256];
    int                 index;

    // Convert this to the PALETTEENTRY structure
    for (index = 0; index < 256; index++) {
      pPaletteEntry[index].peFlags = PC_NOCOLLAPSE|PC_RESERVED;
      pPaletteEntry[index].peRed = 0;
      pPaletteEntry[index].peGreen = 0;
      pPaletteEntry[index].peBlue = 0;
    } /* for */

    // Create the palette and set it.
    hRet = lpDD->CreatePalette(DDPCAPS_8BIT, pPaletteEntry, &lpDDPalette,NULL);
    lpDDSPrimary->SetPalette(lpDDPalette);
  } /* if */

  // Set the background surface and play the FLIC
  okay = Flic->SetData(FLIC_DATA_DDSURFACE, lpDDSBackground);
  if (!okay)
    return InitFail(hwnd, hRet, "FlicSetData (DDSURFACE) FAILED");
  // Set the callback function (faster than a window)
  Flic->SetCallback(EGIcallback);
  Flic->SetParam(FLIC_PARAM_FRNOTIFY, TRUE);
  Flic->Play(hwnd);

  // Show everything
  ShowWindow(hwnd, nCmdShow);
  UpdateWindow(hwnd);
  SetFocus(hwnd);

  return DD_OK;
}

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Initialization, message loop
//-----------------------------------------------------------------------------
int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE /*unused*/,
                   LPSTR /*lpCmdLine*/, int nCmdShow)
{
  MSG msg;

  if (InitApp(hInstance, nCmdShow) != DD_OK)
    return FALSE;

  while (GetMessage(&msg, NULL, 0, 0)) {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  } /* while */
  return msg.wParam;
}
